<link rel="stylesheet" href="//vjs.zencdn.net/7.8.2/video-js.min.css" rel="stylesheet">
<script crossorigin src="https://cdn.jsdelivr.net/gh/josedvq/covfee@8050fa938ae270ae70244c33903ead9d905bca92/covfee/static/js/opencv.js"></script>
<script src="//vjs.zencdn.net/7.8.2/video.min.js"></script>